'	Program :	ParsPerl Module
	Author  :	H. Piske
	Written :	2001-03-10
	Purpose :	find perl string literals and code blocks

	History :

2001-03-20	figure out when / is division and when it is pattern match
	This is far from trivial. The term
	whatever / 4 / +3;
	Is two divisions to perl as long as "whatever" is percieved as a
	constant. If whatever was previously defined as a sub, the above
	line computes m/ 4 /, adds 3 and feeds it to the sub.
	In contrast, this module will always presume that line to be
	a pattern match, assuming nobody will want to divide anything
	into a string constant.
	So, a slash is a division after a closing bracket or a scalar
	or a numerical constant and a pattern match everywhere else.
	Note literals are replaced by scalars, so
	something / foo         starts a pattern match while
	"thisone" / foo         becomes
	$quote[0] / foo         which is a division.

2001-03-21	there can be a space between q (qw,m,s,tr...)
			and the actual delimiter
2001-03-22	numbers (preceeding slash) can consist of more than just digits
		-	some blocks (like map, grep or unnamed blocks) can come up
			with identical names. Provide uniqueness by appending numbers.
		-	=cut can reach thru end of file.
2001-03-24	restrict sub catcher to subs instead of all {} blocks
2001-03-25	mistook -s for s///
';

package ParsPerl;

BEGIN
{
	use Exporter;
	use vars qw(@ISA @EXPORT);
	@ISA = qw(Exporter);
	@EXPORT = qw(ParseStrings ParsePerl ParseBlock);
}
return 1;

=this is how it works

function:	ParseStrings
parameters:	code [, name of array]
returns:	parsed code
	Prepares code for further analysis by placing all strings into vars.
	Recognises '', "", q//, qw//, s/// and so on. Also escapes special
	characters with backslashes.
	Strings are put into an array. If no array name is passed, array
	@quote is used. The array is not cleared, so it will grow with
	more calls to ParseStrings. The returned code is enclosed in double
	quotes, so that an eval will restore the original input.
	Note: do not pass an array reference - just the name (without @).

function:	ParsePerl
parameters:	code [, name of hash]
returns:	parsed code
	Extracts subroutines (enclosed in curlies). Gets confused by curly
	brackets in literals - pass code as prepared by ParseStrings.
	Subs are put into a hash. If no hash name is passed, %block is used.
	The hash is not cleared, so it will grow with subsequent calls.
	Each hash entry is enclosed in double quotes for convenient
	restoring of original code with eval $block{subname};
	Note: do not pass a hash reference - just the name (without %).

function:	ParseBlock
parameters:	code, delimiter [, flag for pre-parsed strings]
returns:	parsed block, remainder of code

=test code
use ParsPerl;
$_ = $INC{'ParsPerl.pm'}; # or any other perl file name
open PERLFILE, $_ or die "ouch";
$code = join '', <PERLFILE>;
close PERLFILE;

$_ = ParsePerl ParseStrings $code;
map {$_ = eval $_} values %block;
print "IT WORX\n" if eval ($_) eq $code;
print 'There are ', $#quote + 1, " strings.\n";
print "$_\n" for '*** subs found:', keys %block;
=cut

sub ParseBlock
{
	local ($header, $code, $_);
	$_ = $_[2] ? $_[0] : ParseStrings ($_[0], _quote_);
	$code = '\\' . $_[1];
	unless (/$code/)
	{
		undef @_quote_;
		return ('', $_[0]);
	}
	$header = $`;
	($code, $_) = block ($', $_[1]);
	$code .= $& if s/^\s*;//;
	unless ($_[2])
	{
		$header = eval ($header.'"');
		$code = eval ('"'.$code.'"');
		$_ = eval ('"'.$_);
		undef @_quote_;
	}
	return ($header . $_[1] . $code, $_);
}
sub ParsePerl
{
	local ($code, $block, $blocks, $snip, $slice, $uni, $name, $_);
	$_ = shift;
	$block = shift || 'block';
	eval ('$blocks = \\%main::' . $block);
	$code = '';
	for (;;)
	{
		#last unless /(\s*)(((\w+):?\s*)+)\{/;		# all {} blocks
		last unless /((\\E.*?\]\s*)*)(\bsub\s+((\w+)\s*)|([A-Z]+)\s*)(\\E.*?\]\s*)?\{/;
		if (substr ($`, -1) eq '$')
		{
			$code .= $` . $&;
			$_ = $';
			next;
		}
		$snip = $1 . $3 . $7;
		$name = $5 || $6;
		$uni = '';
		while (defined $blocks->{$name . $uni})
		{
			$name .= '_' unless $uni++;
		}
		$name .= $uni;
		$code .= "$`\$$block\{'$name'}";
		($slice, $_) = block ($', '{');
		$blocks->{$name} = "\"$snip\{$slice\"";
	}
	return $code . $_;
}
sub block
{
	local ($rem, $del, $end, $cnt, $len, $_);
	($rem, $del) = @_;
	$end = $_ = '\\' . $del;
	y/{[(</}])>/
		?	($end = "[$_$end]")
		:	($del = '');
	$cnt = 1;
	$len = 0;
	$_ = $rem;
	for (;;)
	{
		return ($rem, '') unless /(\\*)($end)/;
		$len += length ($`) + length ($1) + 1;
		$cnt += ($2 eq $del) ? 1 : -1 unless length ($1) & 1;
		last unless $cnt;
		$_ = $';
	}
	return (substr ($rem, 0, $len), substr ($rem, $len));
}
sub ParseStrings
{
	local ($code, $block, $blocks, $del, $rem, $_);
	$_ = shift;
	$block = shift || 'quote';
	($block eq '_quote_')
		?	($blocks = \@_quote_)
		:	eval ('$blocks = \\@main::' . $block);
	#@{$blocks} = ();
	$code = '';
	while (m![/\'\"#\`]|\b(q[qrwx]?|[msy]|tr)\s*\W|^=\w+!m)
	{
		$del = $&;
		$rem = $';
		$_ = $`;
		s/[\$\@\%\\]/\\$&/g;
		$code .= $_;
		if (/\$\s*$/
		|| (	$del =~ /^s/ && substr ($_, -1) eq '-')
		|| (	$del eq '/'
			&& (	$code =~ /[\)\]}]\s*$/
				||  $code =~ /(^|\W)\d+(e\d+)?\s*$/s
				||  $code =~ /[\$\@\%#]\s*\w*\s*$/)))
		{
			$code .= $del;
			$_ = $rem;
			next;
		}
		if ($del =~ /^=/)
		{
			$rem =~ /^=cut/m || $rem =~ /$/s;
			$del .= $` . $&;
			$_ = $';
		}
		elsif ($del eq '#')
		{
			$rem =~ /$/m;
			$del .= $`;
			$_ = $& . $';
			push @{$blocks}, $del;
			$code .= "\\E\$$block\[$#$blocks]";
			next;
		}
		else
		{
			($rem, $_) = block ($rem, substr $del, -1);
			if ($del =~ /(s|y|tr)\s*(.)/)
			{
				$del .= $rem;
				$rem = $2;
				if ($rem =~ /[\(\[{<]/)
				{
					s/(.)//;
					$del .= $rem = $1;
				}
				($rem, $_) = block ($_, $rem);
			}
			$del .= $rem;
			$del .= $& if $del =~ m!^[/mqsyt]! && s/[cdegimosvx]*//;
		}
		push @{$blocks}, $del;
		$code .= "\$$block\[$#$blocks]";
	}
	s/[\$\@\%\\]/\\$&/g;
	return "\"$code$_\"";
}
